package sim.app.sensor_emitter;

import sim.engine.Steppable;
import sim.portrayal.SimplePortrayal2D;
import sim.portrayal.DrawInfo2D;
import sim.util.MutableDouble2D;
import sim.util.Double2D;
import java.awt.geom.Ellipse2D;

/**
 * Agent class that contains all data common to a <code>Sensor</code> or an <code>Emitter</code>
 *
 * @author Adam Campbell
 */
public abstract class Agent extends SimplePortrayal2D implements Steppable{

	/** total number of frequencies that a sensor can sense or an emitter can emit */
	private static int numFrequencies;
	
	/** identifier for agent */
	private int id;
	
	/** <code>x,y</code> coordinates of agent */
	private MutableDouble2D position;

	/** current movement heading of this agent; note, heading is not necessarily related to the direction that an emitter emits its signal */
	private MutableDouble2D heading;

	/** number of units the agent travels when moving */
	private double speed;
	
	/** radius of the physical part of the agent, only affects how the agent is drawn */
	private double physicalRadius;
	
	/** current frequency of this agent */
	private int frequency;

	/** indicates whether or not this agent is disabled */
	private boolean isDisabled;

	/** contains the movement algorithm for this emitter */
	private AgentMover mover;
	
	/**
	 * Constructor.
	 *
	 * @param identifier ID of this Agent
	 * @param initPos initial position of this Agent
	 * @param initHead initial heading of this Agent
	 */
	public Agent(final int identifier, final MutableDouble2D initPos, final MutableDouble2D initHead){

		this.position = new MutableDouble2D(initPos);
		this.heading = new MutableDouble2D(initHead);
		this.id = identifier;

		// set up some default values for each agent
		this.speed = 1.0;
		this.physicalRadius = 5.0;
		this.isDisabled = false;
		this.frequency = identifier%Agent.numFrequencies;
		this.mover = new StationaryMover(); // set up stationary movement as default
	}

	/**
	 * Each agent in the simulation is a circle.
	 * This code was taken from the Ball.java file from Tutorial 5 of MASON.
	 */	
	public boolean hitObject(final Object object, final DrawInfo2D range){

		final double SLOP = 1.0;  // need a little extra diameter to hit circles
		final double width = range.draw.width * this.physicalRadius;
		final double height = range.draw.height * this.physicalRadius*2;

		final Ellipse2D.Double ellipse = new Ellipse2D.Double( 
			range.draw.x-width/2-SLOP, 
			range.draw.y-height/2-SLOP,
			width+SLOP*2,
			height+SLOP*2 );
		return ( ellipse.intersects( range.clip.x, range.clip.y, range.clip.width, range.clip.height ) );
	}

	/**
	 * Moves the agent forward in the direction of its heading.
	 * If the agent is within <code>speed</code> units of the destination point, then place the agent at the destination.
	 *
	 * @param sim the simulation
	 * @param p agent's destination point
	 */
	protected void moveForwardToPoint(final Simulation sim, final Double2D p){

		// check to see if the agent could reach the destination at this time step,
		//   if so, then place the agent at the destination point
		if(this.position.distance(p) <= this.speed){
			this.setPosition(p.x, p.y);
		}else{
			this.position.addIn(this.heading.x * this.speed, this.heading.y * this.speed);
		}
		
		// probably not the best way to do this, but I'm not sure what is
		if(this instanceof Sensor){
			sim.sensorCont2D.setObjectLocation(this, new Double2D(this.getPosition()));
		}else if(this instanceof Emitter){
			sim.emitterCont2D.setObjectLocation(this, new Double2D(this.getPosition()));
		}
	}
	
	/**
	 * Getter for <code>this.id</code>
	 *
	 * @return this agent's id
	 */
	public int getID(){ return this.id; }
	
	/**
	 * Getter for <code>this.position</code>
	 *
	 * @return this agent's position
	 */
	public MutableDouble2D getPosition(){ return this.position; }

	/**
	 * Setter for <code>this.position</code>
	 *
	 * @param p new position for this agent
	 */
	protected void setPosition(final MutableDouble2D p){ this.position = p; }

	/**
	 * Setter for <code>this.position</code>
	 *
	 * @param x x-coordinate
	 * @param y y-coordinate
	 */
	protected void setPosition(final double x, final double y){ this.setPosition(new MutableDouble2D(x, y)); }

	/**
	 * Getter for <code>this.heading</code>
	 *
	 * @return this agent's heading
	 */
	protected MutableDouble2D getHeading(){ return this.heading; }
	
	/**
	 * Setter for <code>this.heading</code>
	 * The new heading should already be normalized.
	 *
	 * @param h new heading
	 */
	protected void setHeading(final MutableDouble2D h){ this.heading = h; }
	
	/**
	 * Setter for <code>this.heading</code>
	 * The new heading should already be normalized.
	 *
	 * @param x x-direction
	 * @param y y-direction
	 */
	protected void setHeading(final double x, final double y){ this.setHeading(new MutableDouble2D(x, y)); }
	
	/**
	 * Getter for <code>this.isDisabled</code>
	 *
	 * @return whether or not this agent is disabled
	 */
	public boolean getIsDisabled(){ return this.isDisabled; }

	/**
	 * Setter for <code>this.isDisabled</code>
	 *
	 * @param b value indicating whether or not this agent is disabled
	 */
	public void setIsDisabled(final boolean b){ this.isDisabled = b; }
	
	/**
	 * Setter for <code>this.physicalRadius</code>
	 *
	 * @param d new physical radius for this agent
	 */
	protected void setPhysicalRadius(final double d){ this.physicalRadius = d; }
		
	/**
	 * Getter for <code>this.physicalRadius</code>
	 *
	 * @return physical radius of this agent
	 */
	protected double getPhysicalRadius(){ return this.physicalRadius; }
	
	/**
	 * Getter for <code>this.frequency</code>
	 *
	 * @return integer value indicating the frequency sensed or emitted by this agent
	 */
	public int getFrequency(){ return this.frequency; }

	/**
	 * Setter for <code>this.frequency</code>
	 *
	 * @param f new frequency
	 */
	public void setFrequency(final int f){ this.frequency = f; }
	
	/**
	 * Getter for <code>this.mover</code>
	 *
	 * @return this agent's AgentMover
	 */
	protected AgentMover getMover(){ return this.mover; }
	
	/**
	 * Setter for <code>this.mover</code>
	 *
	 * @param m new mover
	 */
	protected void setMover(final AgentMover m){ this.mover = m; }

	/**
	 * Getter for <code>Agent.numFrequencies</code>
	 *
	 * @return number of frequencies used by the sensors and emitters
	 */
	protected static int getNumFrequencies(){ return Agent.numFrequencies; }
	
	/**
	 * Setter for <code>Agent.numFrequencies</code>
	 *
	 * @param n number of frequencies used by the sensors and emitters
	 */
	protected static void setNumFrequencies(final int n){ Agent.numFrequencies = n; }
}