package sim.app.sensor_emitter;

/**
 * Used to set up the initial positions of a group of agents.
 * 
 * @author Adam Campbell
 */
class AgentPositioner{

	/**
	 * Returns the number of agents that are to be created by the agent positioner given by the parameter.
	 * No error checking is done in this function.
	 *
	 * @param positionerParams the positioner and additional parameters given for the 'SENSOR_POSITIONER_n' or 'EMITTER_POSITIONER_n' parameter
	 * @param parameterName the name of the parameter passed from the input file, used if warning message is printed out
	 * @return number of agents that should be positioned by the given positioner
	 */
	protected static int numAgentsInPositioner(final String[] positionerParams, final String parameterName){

		// could be implemented using the reflection package, but for now, I'll do the quick and easy if-else way
		if(positionerParams[0].equalsIgnoreCase("SinglePositioner")){
			return 1;
		}else if(positionerParams[0].equalsIgnoreCase("GridPositioner")){
			return Integer.parseInt(positionerParams[1])*Integer.parseInt(positionerParams[2]); // rows*cols for grid; no error checking done here!
		}else{
			System.err.println("(AgentPositioner.numAgentsInPositioner) WARNING: Invalid parameters for " + parameterName);
			return 0;
		}
	}	

	/**
	 * Positions the agents according to the given params.
	 * No error checking is done in this function.
	 *
	 * @param agents list of agents
	 * @param positionerParams the positioner and additional parameters given for the 'SENSOR_POSITIONER_n' or 'EMITTER_POSITIONER_n' parameter
	 * @param parameterName the name of the parameter passed from the input file, used if warning message is printed out
	 */
	protected static void positionAgents(final Agent[] agents, final String[] positionerParams, final String parameterName){

		if(positionerParams[0].equalsIgnoreCase("SinglePositioner")){
			AgentPositioner.singlePositioner(agents, positionerParams);
		}else if(positionerParams[0].equalsIgnoreCase("GridPositioner")){
			AgentPositioner.gridPositioner(agents, positionerParams);
		}else{
			System.err.println("(AgentPositioner.positionAgents) WARNING: Invalid parameters for " + parameterName);
		}
	}
	
	/**
	 * Positions a single agent.
	 * No error checking is done in this function.
	 * 
	 * <code>params</code> should contain exactly three values<br><br>
	 *   (<code>SinglePositioner [x] [y]</code>),<br><br>
	 *   where <code>x y</code> are the coordinates of the agent.
	 *
	 * @param agents should contain one agent
	 * @param params should contain exactly three values (String int int)
	 */
	private static void singlePositioner(final Agent[] agents, final String[] params){
		final double x = Double.parseDouble(params[1]);
		final double y = Double.parseDouble(params[2]);
		
		agents[0].setPosition(x, y);
	}

	/**
	 * Positions a group of agents in a rectangular grid.
	 * No error checking is done in this function.
	 *
	 * <code>params</code> should contain exactly seven values<br><br>
	 *   (<code>GridPositioner [rows] [columns] [x-anchor] [y-anchor] [width] [height]</code>), <br><br>
	 *   where <code>rows columns</code> are the number of rows and columns in the grid,
	 *   <code>x-anchor y-anchor</code> are the upper-left coordinates of the grid,
	 *   and <code>width height</code> are the width and height of the grid.
	 *
	 * @param agents should contain rows*columns number of agents
	 * @param params should contain exactly seven values (String int int double double double double)
	 */
	private static void gridPositioner(final Agent[] agents, final String[] params){
		final int rows = Integer.parseInt(params[1]);
		final int cols = Integer.parseInt(params[2]);
		final double xAnchor = Double.parseDouble(params[3]);
		final double yAnchor = Double.parseDouble(params[4]);
		final double width = Double.parseDouble(params[5]);
		final double height = Double.parseDouble(params[6]);
		final double xSpacing = width/Math.max(1, cols-1);
		final double ySpacing = height/Math.max(1, rows-1);
		
		for(int i = 0, index = 0; i < rows; i++){
			for(int j = 0; j < cols; j++, index++){
				agents[index].setPosition(xAnchor + j*xSpacing, yAnchor + i*ySpacing);
			}
		}
	}
}