package sim.app.sensor_emitter;

import sim.engine.SimState;
import sim.portrayal.DrawInfo2D;
import sim.util.MutableDouble2D;
import java.awt.Graphics2D;
import java.awt.Color;

/**
 * Emitters are one type of agent in the simulation.
 * They emit a signal in some frequency that can be detected by sensors.
 * Emitters do not communicate with each other, but do have algorithms that control their signals.
 *
 * @author Adam Campbell
 */
public class Emitter extends Agent{

	/** colors representing each of the frequencies affect how the emitters are displayed */
	private static Color[] frequencyColors;
	
	/** 
	 * The maximum distance that any emitter is emitting.
	 * Used when obtaining the list of emitters that a sensor can see because
	 *   the emitters that a sensor can see are determined by the emit-distance
	 *   and not by any sense-distance of sensors.
	 * Thus, when using the <code>Continuous2D</code>'s <code>getObjectsWithinDistance()</code>
	 *   function, the distance we need to use is based on the maximum distance that any emitter is emitting.
	 * See the <code>Sensor.obtainSensedEmitters()</code> function for more details.
	 * This value should be maintained in the <code>Simulation</code> class.
	 */
	protected static double MAX_EMIT_DISTANCE;
	
	/** distance that this emitter emit's its signal */
	private double emitDistance;

	/** direction that this emitter is emitting its signal */
	private MutableDouble2D emitDirection;

	/** 
	 * Total number of degrees that this emitter's signal spreads around its <code>emitDirection</code>
	 * That is, the signal spreads to the left <code>emitSpread/2</code> directions to the left and right of the <code>emitDirection</code>.
	 */
	private double emitSpread;
	 
	/** determines whether or not this emitter's signal is drawn */
	private boolean drawEmittedSignal;

	/** contains the signal controlling algorithm for this emitter */
	private SignalController signalController;
		
	/**
	 * Constructor.
	 *
	 * @param identifier ID of this emitter
	 */
	public Emitter(final int identifier){
		super(identifier, new MutableDouble2D(), new MutableDouble2D());
		
		this.emitDistance = 100.0; // arbitrary default value
		this.emitDirection = new MutableDouble2D(1.0, 0.0); // arbitrary default value pointing emitted signal to the right
		this.emitSpread = 10.0; // abritrary default value
		this.drawEmittedSignal = true;
		this.signalController = new StationarySignalController();
	}
	
	/**
	 * Determines what an emitter does during each step of the simulation.
	 */
	public void step(final SimState state){

		// no need to do anything if this emitter is disabled
		if(this.getIsDisabled()) return;
		
		final Simulation sim = (Simulation)state;

		this.getMover().move(this, sim);
		this.signalController.controlSignal(this, sim);
	}

	/**
	 * Draws the emitter as a gray filled circle with a black circle surrounding it.
	 * If <code>drawEmittedSignal</code> is <code>true</code> then the emitter's signal is drawn
	 *   in the color of its frequency.
	 * If the emitter is disabled, then only a black filled circle is drawn.
	 */
	public void draw(final Object object, final Graphics2D graphics, final DrawInfo2D info){
	
		double width, height;
		int x, y, w, h;

		if(this.getIsDisabled()){

			// draw a disabled emitter
			width = info.draw.width * this.getPhysicalRadius();
			height = info.draw.height * this.getPhysicalRadius();
			x = (int)(info.draw.x - width);
			y = (int)(info.draw.y - height);
			w = (int)(width) * 2;
			h = (int)(height) * 2;
			graphics.setColor(Color.black);
			graphics.fillOval(x, y, w, h);
			
		}else{
			// draw this emitter's signal
			if(this.drawEmittedSignal){
				width = info.draw.width * this.emitDistance;
				height = info.draw.height * this.emitDistance;
				x = (int)(info.draw.x - width);
				y = (int)(info.draw.y - height);
				w = (int)(width) * 2;
				h = (int)(height) * 2;
				graphics.setColor(Emitter.frequencyColors[this.getFrequency()]);
		
				final double direction = Math.toDegrees(this.emitDirection.angle()); // angular direction of emitted signal
				final double halfSpread = this.emitSpread/2.0;
				graphics.fillArc(x, y, w, h, (int)Math.round(direction - halfSpread), (int)Math.round(this.emitSpread));
			}
			
			// draw an enabled emitter
			width = info.draw.width * this.getPhysicalRadius();
			height = info.draw.height * this.getPhysicalRadius();
			x = (int)(info.draw.x - width);
			y = (int)(info.draw.y - height);
			w = (int)(width) * 2;
			h = (int)(height) * 2;
			
			// draw the emitter
			graphics.setColor(Color.gray);
			graphics.fillOval(x, y, w, h);

			// draw a black circle around the emitter
			graphics.setColor(Color.black);
			graphics.drawOval(x, y, w, h);
		}
		
	}
	
	/**
	 * Getter for <code>this.emitDistance</code>
	 *
	 * @return the distance that this emitter emits its signal
	 */
	public double getEmitDistance(){ return this.emitDistance; }

	/**
	 * Setter for <code>this.emitDistance</code>
	 *
	 * @param d new emit distance
	 */
	public void setEmitDistance(final double d){ this.emitDistance = d; }

	/**
	 * Getter for <code>this.emitDirection</code>
	 *
	 * @return the vector that indicates where this emitter's signal is pointing
	 */
	public MutableDouble2D getEmitDirection(){ return this.emitDirection; }
	
	/**
	 * Setter for <code>this.emitDirection</code>
	 *
	 * @param d new signal emit-direction
	 */
	public void setEmitDirection(final MutableDouble2D d){ this.emitDirection = d; }

	/**
	 * Getter for <code>this.emitSpread</code>
	 *
	 * @return the angle in degrees 
	 */
	public double getEmitSpread(){ return this.emitSpread; }
	
	/**
	 * Setter for <code>this.emitSpread</code>
	 *
	 * @param d new angle for spread of the signal
	 */
	public void setEmitSpread(final double d){ this.emitSpread = d; }
	
	/**
	 * Setter for <code>this.signalController</code>
	 *
	 * @param sc new signal controller
	 */
	protected void setSignalController(final SignalController sc){ this.signalController = sc; }

	/**
	 * Sets the frequency color array for the Sensor class.
	 *
	 * @param colors color array
	 */
	protected static void setFrequencyColors(final Color[] colors){ Emitter.frequencyColors = colors; }
}
