package sim.app.sensor_emitter;

import sim.util.MutableDouble2D;
import sim.util.Double2D;

/**
 * Moves an agent between a fixed number of waypoints.
 * Cycles through the waypoints.
 * 
 * @author Adam Campbell
 */
class FixedWaypointMover extends AgentMover{

	/** set of waypoints that this agent moves between */
	private Double2D[] waypoints;
	
	/** the current waypoint that the agent is moving towards */
	private int currentWaypoint;
	
	/** 
	 * Constructor that takes in a string array containing the list of waypoints.
	 * The array should contain an odd number of elements (one for 'WaypointMover' and a pair of numbers for every waypoint)
	 * No error checking is done here.
	 *
	 * @param moverParams parameters given from the input file for this mover
	 */
	public FixedWaypointMover(final String[] moverParams){
	
		this.waypoints = new Double2D[moverParams.length/2];
		
		for(int i = 1; i < moverParams.length; i+=2){
			this.waypoints[i/2] = new Double2D(Double.parseDouble(moverParams[i]), Double.parseDouble(moverParams[i+1])); // no error checking!
		}
		
		this.currentWaypoint = 0;
	}
	
	/**
	 * Moves an agent towards its current waypoint.
	 * Once the waypoint is reached, the next waypoint is chosen.
	 * Waypoints cycle through once they are all reached.
	 *
	 * @param a agent to be moved
	 * @param sim the simulation that the agent is in
	 */
	protected void move(final Agent a, final Simulation sim){
	
		// no need to do anything if there's no waypoints
		if(this.waypoints.length == 0) return;
		
		// if we're at the current waypoint, then start moving to the next one
		if(this.waypoints[this.currentWaypoint].equals(a.getPosition())){
			this.currentWaypoint++;
			this.currentWaypoint %= this.waypoints.length; // loop back to 0
		}
		
		// if the agent is already at its current waypoint, then exit out of the function
		// even with the previous check, an agent could be at the current waypoint when there's
		//   only one waypoint or if two consecutive waypoints are at the same location
		if(this.waypoints[this.currentWaypoint].equals(a.getPosition())) return;
		
		// point the agent towards its current waypoint
		final MutableDouble2D h = new MutableDouble2D(this.waypoints[this.currentWaypoint].x - a.getPosition().x, this.waypoints[this.currentWaypoint].y - a.getPosition().y);
		a.setHeading(h.normalize());
		
		a.moveForwardToPoint(sim, this.waypoints[this.currentWaypoint]);
	}
}