package sim.app.sensor_emitter;

import java.util.ArrayList;

/**
 * Implements Method2 from AAMAS 2009 paper (Campbell and Wu).
 * Using the incoming messages from other senors, this sensor counts how many sensors (excluding itself) are monitoring each frequency.
 * One of the frequencies that is the least represented is then randomly selected as the new frequency for this agent. 
 * If the sensor's current frequency is one of the least represented, then it always stays in its current frequency.
 * 
 * @author Adam Campbell
 */
class Method2FrequencyUpdater extends FrequencyUpdater{

	/**
	 * Constructor
	 */
	public Method2FrequencyUpdater(){
	}
	
	/**
	 * Updates the sensor's frequency based on the Method2 algorithm from the AAMAS 2009 paper (Campbell and Wu).
	 *
	 * @param s the sensor
	 * @param sim the simulation that the sensor is in
	 */
	protected void update(final Sensor s, final Simulation sim){
	
		final ArrayList<Integer> lowestCounts = new ArrayList<Integer>();
		final int[] counts = s.frequencyCount(sim.sensorNetwork, false);
		boolean isLeastRepresented = false;
		int min = Integer.MAX_VALUE;
		
		// determine which frequency is the least represented
		for(int n: counts){
			min = Math.min(n, min);
		}
		
		// loop through frequencies to see which one(s) is the least represented
		for(int i = 0; i < counts.length; i++){
			if(counts[i] == min){
				lowestCounts.add(i); // frequency i is least represented
				
				// check to see if the sensor is already monitoring one of the least represented frequencies
				if(s.getFrequency() == i){
					isLeastRepresented = true;
				}
			}
		}
		
		// randomly choose from amongst the least represented frequencies
		if(!isLeastRepresented){
			s.setFrequency(lowestCounts.get(sim.random.nextInt(lowestCounts.size())));
		}
	}
}