package sim.app.sensor_emitter;

import sim.util.Double2D;
import sim.util.MutableDouble2D;

/**
 * Moves an agent to a randomly generated waypoint.
 * The points are generated within the rectangle given by the parameter file
 *
 * @author Adam Campbell
 */
class RandomWaypointMover extends AgentMover{

	/** randomly generated waypoint that the agent is moving towards */
	private Double2D waypoint;

	/** x-coordinate of upper left point of rectangle where randomly generated points are placed */
	private double xAnchor;

	/** y-coordinate of upper left point of rectangle where randomly generated points are placed */
	private double yAnchor;

	/** width of rectangle where randomly generated points are created */
	private double width;

	/** height of rectangle where randomly generated points are created */
	private double height;

	/** 
	 * Constructor that takes in a string array containing the rectangle where points are generated.
	 * The array should contain exactly five elements ('RandomWaypointMover [xAnchor] [yAnchor] [width] [height])
	 * No error checking is done here.
	 *
	 * @param moverParams parameters given from the input file for this mover
	 * @param sim the simulation; used for its random number generator
	 */
	public RandomWaypointMover(final String[] moverParams, final Simulation sim){
	
		this.xAnchor = Double.parseDouble(moverParams[1]);
		this.yAnchor = Double.parseDouble(moverParams[2]);
		this.width = Double.parseDouble(moverParams[3]);
		this.height = Double.parseDouble(moverParams[4]);

		this.waypoint = new Double2D(this.xAnchor + sim.random.nextDouble() * this.width, this.yAnchor + sim.random.nextDouble() * this.height);
	}
	
	/**
	 * Moves an agent to a randomly generated waypoint.
	 * Once that waypoint is reached, a new waypoint is randomly generated.
	 *
	 * @param a agent to be moved
	 * @param sim the simulation that the agent is in
	 */
	protected void move(final Agent a, final Simulation sim){
		
		// if we're at the waypoint, then generate a new one
		if(this.waypoint.equals(a.getPosition())){
			this.waypoint = new Double2D(this.xAnchor + sim.random.nextDouble() * this.width, this.yAnchor + sim.random.nextDouble() * this.height);
		}
		
		// if the agent is already at its current waypoint, then exit out of the function
		// even with the previous check, an agent could be at the current waypoint if it just so happens
		//   that the newly generated waypoint is equal to the last one
		// this is unlikely to occur, but we don't want to normalize a zero-length vector
		if(this.waypoint.equals(a.getPosition())) return;
		
		// point the agent towards its current waypoint
		final MutableDouble2D h = new MutableDouble2D(this.waypoint.x - a.getPosition().x, this.waypoint.y - a.getPosition().y);
		a.setHeading(h.normalize());
		
		a.moveForwardToPoint(sim, this.waypoint);
	}
}