package sim.app.sensor_emitter;

import java.io.PrintWriter;
import java.util.Arrays;
import sim.util_amc.ConfigParams;
import sim.util_amc.DataGatherer;
import sim.engine.SimState;

/*import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;*/

/**
 * SenseEmitDataGatherer contains all of the file writers for data collection during simulation runs.
 *
 * @author Adam Campbell
 */
class SenseEmitDataGatherer extends DataGatherer{
	
	/** writes out the sensors that are sensing an emitter */
	private PrintWriter sensingSensorsPW;

	/** writes out the number of sensors that are sensing each emitter */
	private PrintWriter sensedCountPW;
	
	/** keeps track of the number of sensors that are sensing each emitter */
	private int[] numSensed;
	
	/**
	 * Constructor.  Opens all of the output files for writing.
	 *
	 * @param p parameter file for the current run
	 * @param o output directory
	 * @param seed seed used for the random number generator
	 */
	public SenseEmitDataGatherer(final ConfigParams p, final String o, final long seed){
		super(p, o, seed);
		
		this.sensingSensorsPW = super.openFileToWrite("sensingSensors");
		this.sensingSensorsPW.println("TIME SENSOR_ID EMITTER_ID FREQUENCY");
		
		this.sensedCountPW = super.openFileToWrite("sensedCount");
		this.sensedCountPW.println("TIME EMITTER_0 EMITTER_1 EMITTER_2 ...");
		
		this.numSensed = null; // eventually will be set in printSensingSensors()
	}
	
	/**
	 * Writes to the output files.  Function is called by the simulation's scheduler.
	 *
	 * @param state the simulation
	 */
	public void step(final SimState state){

		final Simulation sim = (Simulation)state;
		
		this.printSensingSensors(sim);
		this.printSensedCount(sim);
	}
	
	/**
	 * For each sensor, print out the scheduler's current time, the sensor's ID, the ID of any 
	 *   emitter that the sensor is sensing, and the frequency of the sensor/emitter.
	 * Function also computes the number of sensors that are sensing each emitter, as this information
	 *   is used in printSensedCount().
	 *
	 * @param sim the simulation
	 */
	private void printSensingSensors(final Simulation sim){
	
		// should only occur one time
		if(this.numSensed == null) this.numSensed = new int[sim.getNumberOfEmitters()];
		
		Arrays.fill(this.numSensed, 0);
		
		for(Object o1: sim.sensorBag){
			Sensor s = (Sensor)o1;

			for(Object o2: s.sensedEmitters){
				Emitter e = (Emitter)o2;
				this.sensingSensorsPW.println(sim.schedule.getSteps() + " " + s.getID() + " " + e.getID() + " " + e.getFrequency());
				this.numSensed[e.getID()]++;
			}
		}
	}

	/**
	 * For each time step, print out the number of sensors that are sensing each emitter.
	 * Each line contains a number for every emitter, even if the emitter is not being sensed.
	 *
	 * @param sim the simulation
	 */
	private void printSensedCount(final Simulation sim){
	
		this.sensedCountPW.print(sim.schedule.getSteps());
		
		for(int n: this.numSensed){
			this.sensedCountPW.print(" " + n);
		}
			
		this.sensedCountPW.println();
	}
}