package sim.app.sensor_emitter;

import sim.engine.SimState;
import sim.portrayal.DrawInfo2D;
import sim.field.network.Network;
import sim.field.network.Edge;
import sim.util.Bag;
import sim.util.Double2D;
import sim.util.MutableDouble2D;
import sim.util_amc.MathFunctions;
import java.awt.Graphics2D;
import java.awt.Color;

/**
 * Sensors are one type of agent in the simulation.
 * They sense emitters and communicate their currently sensed frequency with other sensors.
 * As of now, sensors are omni-directional
 *
 * @author Adam Campbell
 */
public class Sensor extends Agent{

	/** colors representing each of the frequencies affect how the sensors are displayed */
	private static Color[] frequencyColors;
	
	/** the distance that a sensor communicates its information to neighboring sensors */
	private double commDistance;
	
	/** indicates whether or not this sensor is detecting an emitter */
	private boolean isDetectingEmitter;
	
	/** the probability that a sensor update its frequency during each time step */
	private double updateProbability;
	
	/** algorithm used to control which frequency this sensor senses */
	private FrequencyUpdater frequencyUpdater;
	
	/** Bag containing the emitters that this sensor is currently sensing.  Used primarily for data collection purposes. */
	protected Bag sensedEmitters;
	
	/**
	 * Constructor.
	 *
	 * @param identifier ID of this sensor
	 */
	public Sensor(final int identifier){
		super(identifier, new MutableDouble2D(), new MutableDouble2D());
		
		this.commDistance = 100.0; // arbitrary default value
		this.updateProbability = 1.0;
		this.isDetectingEmitter = false;
		this.frequencyUpdater = new StationaryFrequencyUpdater(); // arbitrary default value

		this.sensedEmitters = new Bag();
		
		this.setFrequency(0);
	}
	
	/**
	 * Sensors perform the following actions each time step if it is enabled:
	 * <li>
	 *   <ul> Update which frequency they are sensing
	 *   <ul> Clear any messages that they currently have
	 *   <ul> Move to a new location
	 *   <ul> Sense emitters
	 *   <ul> Broadcast information to neighboring sensors
	 * </li>
	 *
	 * If the sensor is disabled, then only clear its messages.
	 */
	public void step(final SimState state){

		final Simulation sim = (Simulation)state;

		if(this.getIsDisabled()){

			// clear incoming messages
			this.clearIncomingMessages(sim.sensorNetwork);
		}else{

			// update the frequency
			if(sim.random.nextDouble() < this.updateProbability){
				this.frequencyUpdater.update(this, sim);
			}

			// clear incoming messages
			this.clearIncomingMessages(sim.sensorNetwork);

			// move
			this.getMover().move(this, sim);
		
			// sense nearby emitters
			this.obtainSensedEmitters(sim);
			this.isDetectingEmitter = (sensedEmitters.size() > 0);
		
			// broadcast to neighboring sensors
			this.broadcastInformation(sim);
		}
	}

	/**
	 * Removes all of the sensor networks edges coming into this sensor
	 *
	 * @param sensorNetwork the sensor network
	 */
	private void clearIncomingMessages(final Network sensorNetwork){

		// can't directly iterate through incomingEdges because this is a reference to the bag used
		//   in the Network class
		final Bag incomingEdges = sensorNetwork.getEdgesIn(this);
		final Bag copyOfIncomingEdges = new Bag();
		
		// make a copy
		for(Object o: incomingEdges) copyOfIncomingEdges.add(o);
		
		for(Object o: copyOfIncomingEdges){
			sensorNetwork.removeEdge((Edge)o);
		}
	}
	
	/**
	 * Fills <code>this.sensedEmitters</code> with all of the emitters that are sensed by this sensor.
	 * An emitter is sensed when the following criteria are met: 
	 *   1) the emitter is enabled,
	 *   2) the emitter's emit-distance is greater than or equal to the distance between the emitter and this sensor,
	 *   3) the sensor falls within the emitter's transmit area, and
	 *   4) the sensor is sensing the same frequency that the emitter is emitting.
	 *
	 * @param sim the simulation
	 */
	private void obtainSensedEmitters(final Simulation sim){
	
		final Bag potentialNeighbors = sim.emitterCont2D.getObjectsWithinDistance(new Double2D(this.getPosition()), Emitter.MAX_EMIT_DISTANCE);
		
		this.sensedEmitters.clear();
		
		// check to make sure the emitter satisfies the criteria listed in the comments above
		for(Object o: potentialNeighbors){
			final Emitter e = (Emitter)o;
			
			// obtain angle between emitter's signal direction and vector between emitter and sensor
			double angle;
			if(this.getPosition().equals(e.getPosition())){
				angle = 0.0;
			}else{
				// negate the y-value because the y-axis in java-apps has positive value facing down
				final MutableDouble2D vectorEmitterToSensor = (new MutableDouble2D(this.getPosition().x-e.getPosition().x, -(this.getPosition().y-e.getPosition().y))).normalize();
				angle = Math.abs(MathFunctions.angleBetween(vectorEmitterToSensor, e.getEmitDirection()));
			}
			
			if(!e.getIsDisabled() &&
			   this.getPosition().distance(e.getPosition()) <= e.getEmitDistance() && 
			   angle <= e.getEmitSpread() / 2.0 && 
			   this.getFrequency() == e.getFrequency()){
				this.sensedEmitters.add(e);
			}
		}
	}
	
	/**
	 * Broadcasts messages to all sensors within this sensor's communication range.
	 *
	 * @param sim the simulation
	 */
	private void broadcastInformation(final Simulation sim){

		final Bag potentialNeighbors = sim.sensorCont2D.getObjectsWithinDistance(new Double2D(this.getPosition()), this.commDistance);
		final Bag actualNeighbors = new Bag();
		
		// check to make sure the neighboring sensor is within this sensor's communication range
		// also, don't send a message to yourself or to disabled sensors
		for(Object o: potentialNeighbors){
			final Sensor s = (Sensor)o;
			if(this.getID() != s.getID() && this.getPosition().distance(s.getPosition()) <= this.commDistance && !s.getIsDisabled()){
				actualNeighbors.add(o);
			}
		}
		
		for(Object o: actualNeighbors){
			final Sensor s = (Sensor)o;
			sim.sensorNetwork.addEdge(this, s, new Message(this.getID(), sim.schedule.getSteps(), this.getFrequency())); // TODO: send correct message here
		}
	}
	
	/**
	 * Function computes the frequency counts based on the messages that this sensor has received.
	 *
	 * @param sensorNetwork the sensor network
	 * @param includeSelf indicates whether or not this sensor's frequency should be used when obtaining the count
	 * @return array containing the frequency counts for this sensor's neighbors (and possibly itself)
	 */
	protected int[] frequencyCount(final Network sensorNetwork, final boolean includeSelf){
	
		final int[] counts = new int[Agent.getNumFrequencies()];
		final Bag incomingEdges = sensorNetwork.getEdgesIn(this);
		
		// obtain the counts
		for(Object o: incomingEdges){
			final Edge e = (Edge)o;
			final Message m = (Message)e.info;
			counts[m.frequency]++;
		}
		
		// count self?
		if(includeSelf){
			counts[this.getFrequency()]++;
		}
		
		return counts;
	}
	
	/**
	 * Draws the agent
	 */
	public void draw(final Object object, final Graphics2D graphics, final DrawInfo2D info){
		
		double width, height;
		int x, y, w, h;

		if(this.getIsDisabled()){

			// draw a disabled emitter
			width = info.draw.width * this.getPhysicalRadius();
			height = info.draw.height * this.getPhysicalRadius();
			x = (int)(info.draw.x - width);
			y = (int)(info.draw.y - height);
			w = (int)(width) * 2;
			h = (int)(height) * 2;	
			graphics.setColor(Color.black);
			graphics.fillOval(x, y, w, h);
		
		}else{
		
			// draw a disabled emitter
			width = info.draw.width * this.getPhysicalRadius();
			height = info.draw.height * this.getPhysicalRadius();
			x = (int)(info.draw.x - width);
			y = (int)(info.draw.y - height);
			w = (int)(width) * 2;
			h = (int)(height) * 2;
			graphics.setColor(Sensor.frequencyColors[this.getFrequency()]);
			graphics.fillOval(x, y, w, h);
			
			// draw a black outline around the sensor
			graphics.setColor(Color.black);
			graphics.drawOval(x, y, w, h);
				
			// allow the user to see that the agent is detecting an emitter
			if(this.isDetectingEmitter){
				width = info.draw.width * 15;
				height = info.draw.height * 15;
				x = (int)(info.draw.x - width);
				y = (int)(info.draw.y - height);
				w = (int)(width) * 2;
				h = (int)(height) * 2;
				graphics.setColor(Color.gray);
				graphics.drawOval(x, y, w, h);
			}
		}
	}

	/**
	 * Getter for <code>this.updateProbability</code>.
	 *
	 * @return this sensor's frequency update probability
	 */
	public double getUpdateProbability(){ return this.updateProbability; }
	
	/**
	 * Setter for <code>this.updateProbability</code>
	 *
	 * @param p new update probability
	 */
	public void setUpdateProbability(final double p){ this.updateProbability = p; }

	/**
	 * Getter for <code>this.commDistance</code>.
	 *
	 * @return the distance this sensor communicates its information
	 */
	public double getCommDistance(){ return this.commDistance; }

	/**
	 * Setter for <code>this.commDistance</code>
	 *
	 * @param d new communication distance for this sensor
	 */
	public void setCommDistance(final double d){ this.commDistance = d; }

	/**
	 * Setter for <code>this.frequencyUpdater</code>.
	 *
	 * @param fu new frequency updater
	 */
	protected void setFrequencyUpdater(final FrequencyUpdater fu){ this.frequencyUpdater = fu; }
	
	/**
	 * Sets the frequency color array for the Sensor class.
	 *
	 * @param colors color array
	 */
	protected static void setFrequencyColors(final Color[] colors){ Sensor.frequencyColors = colors; }
}
