package sim.util_amc;

import sim.engine.Steppable;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;

/**
 * DataGatherer contains the basic functionality to create data gathering code for a simulation run.<br><br>
 *
 * It performs the following tasks:
 * <li>
 *   <ul> creates an output directory if one does not exist
 *   <ul> obtains and updates (or creates) the <code>run.num</code> file in the given directory
 *   <ul> prints out the meta data for this run (i.e., RNG seed and time that run was started)
 *   <ul> prints a copy of the parameter file
 * </li>
 *
 * Implements <code>Steppable</code> because a data gatherer should be scheduled as data may need to be written each time step of the simulation.
 * Note, if <code>WRITE_DATE: false</code> is set in the <code>run.conf</code> file, then this class will do nothing.
 *
 * @author Adam Campbell
 */
public abstract class DataGatherer implements Steppable{

	/** run number for the current run */
	private int runNum;

	/** directory that all of the files for the current run are written to */
	private String outputDir;

	/** keeps track of all the files that have been opened by <code>openFileToWrite()</code> */
	private ArrayList<PrintWriter> openFiles;
	
	/**
	 * constructor which does tasks outlined in class description
	 *
	 * @param p parameter file for the current run
	 * @param o output directory
	 * @param seed seed used for the random number generator
	 */
	public DataGatherer(final ConfigParams p, final String o, final long seed){

		this.outputDir = o;
		this.makeDirectory();
		this.runNum = this.obtainAndUpdateRunNum();
		this.printMetaData(seed);
		this.openFiles = new ArrayList<PrintWriter>();
		p.printToFile(this.outputDir + "/run" + this.runNum + ".params");
	}

	/**
	 * Opens up a <code>PrintWriter</code> for a file in this gatherer's output directory, and names the file
	 *   'run[runNum].[extension]', where runNum is <code>this.runNum</code> and extension is the value given
	 *   as the parameter to this function.
	 * Also adds the newly created <code>PrintWriter</code> to <code>this.openFiles</code> so it can be closed
	 *   later in <code>finish()</code>.
	 *
	 * @param extension used to name the file
	 * @return <code>PrintWriter</code> for the file, or <code>null</code> if the file couldn't be opened for writing
	 */
	public PrintWriter openFileToWrite(final String extension){

		try{
			final PrintWriter pr = new PrintWriter(new BufferedWriter(new FileWriter(this.outputDir + "/run" + this.runNum + "." + extension)));
			this.openFiles.add(pr);
			return pr;
		}catch(IOException e){
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Should be called whenever a simulation has completed.
	 * Closes all of the files in <code>this.openFiles</code>.
	 */
	public void finish(){
	
		for(PrintWriter pr: this.openFiles){
			pr.close();
		}
	}
	
	/**
	 * Called by the constructor to create the output directory.
	 */
	private void makeDirectory(){
		if(!(new File(this.outputDir)).exists()){
			(new File(this.outputDir)).mkdirs();
		}
	}

	/**
	 * Called by the constructor to print out RNG seed and time started for the current run.
	 *
	 * @param seed seed used for the random number generator
	 */
	private void printMetaData(final long seed){

		try{
			final PrintWriter pr = new PrintWriter(new BufferedWriter(new FileWriter(this.outputDir + "/run" + this.runNum + ".meta")));
			pr.println("RNG seed: " + seed);
			pr.println("Date ran: " + System.currentTimeMillis());
			pr.close();
		}catch(IOException e){
			e.printStackTrace();
		}
	}

	/**
	 * Called by the constructor to determine the current run number.
	 * The run number is obtained by looking at the <code>run.num</code> file in the output directory.
	 * If <code>run.num</code> does not exist, the run number is assumed to be zero.
	 * This function writes out the current run number plus one to <code>run.num</code>.
	 * (Or, writes a <code>1</code> if <code>run.num</code> didn't already exist.)
	 *
	 * @return current run number
	 */
	private int obtainAndUpdateRunNum(){

		int num;

		// obtain run number from run.num if it exists
		// if run.num does not exist, then the current run number is zero
		try{	
			final BufferedReader br = new BufferedReader(new FileReader(this.outputDir + "/run.num"));
			num = Integer.parseInt(br.readLine().trim());
			br.close();
		}catch(IOException e){
			num = 0;
		}

		// write out the new run number to the run.num file
		try{
			final PrintWriter runNumWriter = new PrintWriter(new BufferedWriter(new FileWriter(this.outputDir + "/run.num")));
			runNumWriter.println(num+1);
			runNumWriter.close();
		}catch(IOException e){
			e.printStackTrace();
		}

		return num;
	}
}