package sim.app.sensor_emitter;

import sim.engine.SimState;
import sim.engine.Schedule;
import sim.field.continuous.Continuous2D;
import sim.field.network.Network;
import sim.util.Bag;
import sim.util.Double2D;
import sim.util.MutableDouble2D;
import sim.util_amc.DataGatherer;
import ec.util.MersenneTwisterFast;

/**
 * Contains all of the information for the simulation.
 * This is the main class file for the simulation.
 *
 * @author Adam Campbell
 */
public class Simulation extends SimState{

	/** used for various Java reflection purposes */
	final String packageName = "sim.app.sensor_emitter"; 

	/** Width of the simulation environment */
	private int fieldWidth;
	
	/** Height of the simulation environment */
	private int fieldHeight;

	/** Keeps track of all the sensors. */
	protected Continuous2D sensorCont2D;

	/** Keeps track of all the sensors. */
	protected Continuous2D emitterCont2D;

	/** keeps track of all directed communication links between sensors */
	protected Network sensorNetwork;
	
	/** configuration parameters for this simulation */
	private SenseEmitConfigParams params;
	
	/** data gatherer for this simulation */
	private SenseEmitDataGatherer dataGatherer;
	
	/** random seed for this simulation */
	private long seed;

	/** Bag of this simulation's sensors.  Used mainly when writing data about the simulation. */
	protected Bag sensorBag = new Bag();

	/** Bag of this simulation's emitters.  Used mainly when writing data about the simulation. */
	protected Bag emitterBag = new Bag();

	public static void main(String[] args){
		doLoop(Simulation.class, args);
		System.exit(0);
	}

	/**
	 * Constructor
	 *
	 * @param s RNG seed for this simulation
	 * @param paramFileName parameter file name for this simulation
	 */
	public Simulation(final long s, final String paramFileName){

		super(new MersenneTwisterFast(s), new Schedule());

		this.seed = s;
		this.params = new SenseEmitConfigParams(paramFileName);
	}
	
	/**
	 * Constructor
	 *
	 * @param s RNG seed for this simulation
	 */
	public Simulation(long s){
		this(s, "run.conf");
	}

	/**
	 * this function gets called each time a new simulation is started
	 */
	public void start(){
		super.start();
		
		this.sensorBag.clear();
		this.emitterBag.clear();

		this.createFields();
		
		Agent.setNumFrequencies(this.params.paramInt("NUM_FREQUENCIES"));
		Sensor.setFrequencyColors(this.params.sensorColorParam());
		Emitter.setFrequencyColors(this.params.emitterColorParam());
		this.createSensors();
		this.createEmitters();
		this.startDataGatherer();
	}

	/**
	 * Creates all of <code>Continuous2D</code> objects and <code>Network</code>
	 *   for this simulation.
	 */
	private void createFields(){

		this.fieldWidth = this.params.paramInt("FIELD_WIDTH");
		this.fieldHeight = this.params.paramInt("FIELD_HEIGHT");
		
		this.sensorCont2D = new Continuous2D(this.params.paramDouble("SENSOR_DISCRETIZATION"), this.fieldWidth, this.fieldHeight);
		this.emitterCont2D = new Continuous2D(this.params.paramDouble("EMITTER_DISCRETIZATION"), this.fieldWidth, this.fieldHeight);
		this.sensorNetwork = new Network(true);
	}

	/**
	 * If <code>WRITE_DATA=true</code> from the parameter file, then this function 
	 *   creates and schedules the data gatherer.
	 * Otherwise, this function does nothing.
	 */
	private void startDataGatherer(){

		if(this.params.paramBoolean("WRITE_DATA")){
			this.dataGatherer = new SenseEmitDataGatherer(this.params, this.params.paramString("OUTPUT_DIR"), this.seed);
			
			// parameters are: time, schedule order, event, interval
			this.schedule.scheduleRepeating(0.0, 1, this.dataGatherer, this.params.paramInt("DATA_WRITE_INTERVAL")); 
		}
	}
	
	/**
	 * Creates the sensors that are in the simulation.
	 * The created sensors are those specified by the simulation's parameter file.
	 */
	private void createSensors(){
	
		final int numSensorGroups = this.params.paramInt("NUM_SENSOR_GROUPS"); // sensor groups are numbered 1 to numSensorGroups, inclusive

		for(int i = 1; i <= numSensorGroups; i++){
			final Sensor[] tmpSensors = this.processSensorGroup(i);

			for(Sensor s: tmpSensors){
				this.sensorCont2D.setObjectLocation(s, new Double2D(s.getPosition()));
				this.sensorNetwork.addNode(s);
				
				// parameters are: time, schedule order, event, interval
				this.schedule.scheduleRepeating(0.0, 0, s, 1);
			}
		}
		
	}

	/**
	 * Creates the emitters that are in the simulation.
	 * The created emitters are those specified by the simulation's parameter file.
	 */
	private void createEmitters(){
	
		// create all of the emitters
		// emitter groups are numbered 1 to numSensorGroups, inclusive
		final int numEmitterGroups = this.params.paramInt("NUM_EMITTER_GROUPS");

		for(int i = 1; i <= numEmitterGroups; i++){
			final Emitter[] tmpEmitters = this.processEmitterGroup(i);

			for(Emitter e: tmpEmitters){
				this.emitterCont2D.setObjectLocation(e, new Double2D(e.getPosition()));

				// parameters are: time, schedule order, event, interval
				this.schedule.scheduleRepeating(0.0, 0, e, 1);
				
				// make sure to maintain this value
				Emitter.MAX_EMIT_DISTANCE = Math.max(Emitter.MAX_EMIT_DISTANCE, e.getEmitDistance());
			}
		}
	}
	
	/**
	 * Creates a group of sensors.
	 *
	 * @param num number of the group 
	 * @return the array of sensors that are created
	 */
	private Sensor[] processSensorGroup(final int num){

		final String[] positionerParams = this.params.paramStringArray("SENSOR_POSITIONER_" + num);
		final String[] assignerParams = this.params.paramStringArray("SENSOR_INITIAL_FREQUENCY_ASSIGNER_" + num);
		final String[] moverParams = this.params.paramStringArray("SENSOR_MOVER_" + num);
		final String[] frequencyUpdaterParams = this.params.paramStringArray("SENSOR_FREQUENCY_UPDATER_" + num);
		final double pRadius = this.params.paramDouble("SENSOR_PHYSICAL_RADIUS_" + num); // physical radius determining how big each sensor is drawn
		final double updateProb = this.params.paramDouble("SENSOR_UPDATE_PROBABILITY_" + num); // probability that a sensor update its frequency each time step
		final double commDist = this.params.paramDouble("SENSOR_COMM_DISTANCE_" + num); // distance that sensors broadcast their information
		
		// positioner parameters is invalid, so print warning message and return an empty array 
		// note, moverParams isn't checked as it's an optional parameter; if it's omitted, we assume the sensors are stationary
		// same with the frequencyUpdaterParams
		if(positionerParams.length < 1){
			System.err.println("(Simulation.processSensorCreator) WARNING: Invalid parameters for 'SENSOR_POSITIONER_" + num + "'");
			return new Sensor[]{};
		}

		// number of sensors is determined by the positioner
		final Sensor[] sensors = new Sensor[AgentPositioner.numAgentsInPositioner(positionerParams, "SENSOR_POSITIONER_" + num)];
		for(int i = 0; i < sensors.length; i++){
			sensors[i] = new Sensor(this.sensorBag.size()); // this.numSensors is used as the Sensor's ID
			sensors[i].setPhysicalRadius(pRadius);
			sensors[i].setUpdateProbability(updateProb);
			sensors[i].setCommDistance(commDist);
			sensors[i].setMover(this.determineMover(moverParams));
			sensors[i].setFrequencyUpdater(this.determineFrequencyUpdater(frequencyUpdaterParams));
			this.sensorBag.add(sensors[i]);
		}
		AgentPositioner.positionAgents(sensors, positionerParams, "SENSOR_POSITIONER_" + num);
		InitialFrequencyAssigner.assignFrequencies(sensors, assignerParams, "SENSOR_INITIAL_FREQUENCY_ASSIGNER_" + num, this.random);

		return sensors;
	}

	/**
	 * Creates a group of emitters, positioned using the algorithm specified by EMITTER_POSITIONER_num in the parameter file.
	 *
	 * @param num number of the group 
	 * @return the array of emitter that are created
	 */
	private Emitter[] processEmitterGroup(final int num){

		final String[] positionerParams = this.params.paramStringArray("EMITTER_POSITIONER_" + num);
		final String[] assignerParams = this.params.paramStringArray("EMITTER_INITIAL_FREQUENCY_ASSIGNER_" + num);
		final String[] moverParams = this.params.paramStringArray("EMITTER_MOVER_" + num);
		final String[] signalControllerParams = this.params.paramStringArray("EMITTER_SIGNAL_CONTROLLER_" + num);
		final double[] signalInfoParams = this.params.paramDoubleArray("EMITTER_SIGNAL_INFO_" + num);
		final double pRadius = this.params.paramDouble("EMITTER_PHYSICAL_RADIUS_" + num); // physical radius determining how big each sensor is drawn
		
		// positionerParams was invalid, so print warning message and return an empty array 
		// note, moverParams isn't checked as it's an optional parameter; if it's omitted, we assume the emitters are stationary
		if(positionerParams.length < 1){
			System.err.println("(Simulation.processEmitterGroup) WARNING: Invalid parameters for 'EMITTER_POSITIONER_" + num + "'");
			return new Emitter[]{};
		}
		
		// number of emitters is determined by the positioner
		final Emitter[] emitters = new Emitter[AgentPositioner.numAgentsInPositioner(positionerParams, "EMITTER_POSITIONER_" + num)];	
		for(int i = 0; i < emitters.length; i++){
			emitters[i] = new Emitter(this.emitterBag.size()); // this.numSensors is used as the Emitter's ID
			emitters[i].setPhysicalRadius(pRadius);
			
			// only set the signal settings if there were four double values given for EMITTER_SIGNAL_INFO_num
			if(signalInfoParams.length == 4){
				emitters[i].setEmitDirection(new MutableDouble2D(signalInfoParams[0], signalInfoParams[1]));
				emitters[i].setEmitDistance(signalInfoParams[2]);
				emitters[i].setEmitSpread(signalInfoParams[3]);
			}
			
			emitters[i].setMover(this.determineMover(moverParams));
			emitters[i].setSignalController(this.determineSignalController(signalControllerParams));
			
			this.emitterBag.add(emitters[i]);
		}
		AgentPositioner.positionAgents(emitters, positionerParams, "EMITTER_POSITIONER_" + num);
		InitialFrequencyAssigner.assignFrequencies(emitters, assignerParams, "EMITTER_INITIAL_FREQUENCY_ASSIGNER_" + num, this.random);
		
		return emitters;
	}

	/**
	 * returns a signal controller as specified by the given parameters from the parameter file
	 *
	 * @param moverParams parameters given for some '[SENSOR|EMITTER]_SIGNAL_CONTROLLER_n'
	 * @return AgentMover used by a sensor or emitter
	 */
	private AgentMover determineMover(final String[] moverParams){
	
		if(moverParams.length == 0){ // parameter was not given, so go with default option
			return new StationaryMover();
		}else if(moverParams[0].equalsIgnoreCase("StationaryMover")){
			return new StationaryMover();
		}else if(moverParams[0].equalsIgnoreCase("FixedWaypointMover")){
			return new FixedWaypointMover(moverParams);
		}else if(moverParams[0].equalsIgnoreCase("RandomWaypointMover")){
			return new RandomWaypointMover(moverParams, this);
		}else{ // parameter wasn't one of the above, so go with default option
			return new StationaryMover();
		}	
	}

	/**
	 * returns a frequency updater as specified by the given parameters from the parameter file
	 *
	 * @param frequencyUpdaterParams parameters given for some 'EMITTER_SIGNAL_CONTROLLER_n'
	 * @return FrequencyUpdater used by a sensor
	 */
	private FrequencyUpdater determineFrequencyUpdater(final String[] frequencyUpdaterParams){
	
		if(frequencyUpdaterParams.length == 0){ // parameter was not given, so go with default option
			return new StationaryFrequencyUpdater();
		}else if(frequencyUpdaterParams[0].equalsIgnoreCase("Method1FrequencyUpdater")){
			return new Method1FrequencyUpdater();
		}else if(frequencyUpdaterParams[0].equalsIgnoreCase("Method2FrequencyUpdater")){
			return new Method2FrequencyUpdater();
		}else if(frequencyUpdaterParams[0].equalsIgnoreCase("Method3FrequencyUpdater")){
			return new Method3FrequencyUpdater();
		}else{ // parameter wasn't one of the above, so go with default option
			return new StationaryFrequencyUpdater();
		}	
	}
	
	/**
	 * returns a signal controller as specified by the given parameters from the parameter file
	 *
	 * @param signalControllerParams parameters given for some 'EMITTER_SIGNAL_CONTROLLER_n'
	 * @return SignalController used by an emitter
	 */
	private SignalController determineSignalController(final String[] signalControllerParams){
	
		if(signalControllerParams.length == 0){ // parameter was not given, so go with default option
			return new StationarySignalController();
		}else if(signalControllerParams[0].equalsIgnoreCase("StationarySignalController")){
			return new StationarySignalController();
		}else if(signalControllerParams[0].equalsIgnoreCase("FixedPointSignalController")){
			return new FixedPointSignalController(signalControllerParams);
		}else{ // parameter wasn't one of the above, so go with default option
			return new StationarySignalController();
		}	
	}
	
	/**
	 * Called when the simulation stops
	 */
	public void finish(){
		super.finish();
		
		if(this.params.paramBoolean("WRITE_DATA")){
			this.dataGatherer.finish();
		}
	}

	/**
	 * Returns the number of sensors in the simulation.
	 *
	 * @return number of sensors
	 */
	public int getNumberOfSensors(){ return this.sensorBag.size(); }

	/**
	 * Returns the number of emitters in the simulation.
	 *
	 * @return number of sensors
	 */
	public int getNumberOfEmitters(){ return this.emitterBag.size(); }
}
