package sim.util_amc;

import java.util.StringTokenizer;
import java.util.Hashtable;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import sim.util.Double2D;
import java.awt.Color;

/**
 * ConfigParams takes in a run's parameter file name and creates a <code>Hashtable</code> mapping parameter names to their values.
 * Functions for retrieving typed parameters are given (e.g., <code>paramInt</code>, <code>paramLong</code>, etc.).
 * Comments can be put in a parameter file by starting a line with #.
 * Parameter names are case insensitive.
 * If two parameter names are the same, the one that comes latest in the file will be used.
 * Parameters should follow the format <code>name: value</code>.
 * Each parameter must be on a separate line.
 *
 * @author Adam Campbell
 */
public class ConfigParams{

	/** Maps parameter names (of type <code>String</code>) to their values (also of type <code>String</code>).
	 *  Use proper <code>get</code> function to obtain correctly typed parameter
	 */
	private Hashtable<String, String> table;

	/** A line-by-line copy of the text in the parameter file.  Used by <code>printToFile</code> to write out the file. */
	private String completeFile;

	/**
	 * Constructor that takes in parameter file name, builds hash table mapping parameter names to their values, and constructs the <code>completeFile</code> string.
	 *
	 * @param fileName name of parameter file that is to be read in.
	 */
	public ConfigParams(final String fileName){

		BufferedReader br = null;
		String line;
		String key, value;

		this.table = new Hashtable<String, String>();
		this.completeFile = null;

		// If there is an error opening the parameter file, print out an error and
		//   exit the constructor
		try{
			br = new BufferedReader(new FileReader(fileName));
		}catch(IOException e){
			System.err.println("(ConfigParams.ConfigParams) WARNING: Could not open parameter file: " + fileName);
			return;
		}

		// read in the file
		try{

			// Read in all of the parameters, and put the key/value pair in the hash table
			// We will allow for comments in the parameter file
			//   - Comments in the parameter file will be ignored
			//   - Comments are lines that begin with a '#'
			while((line = br.readLine()) != null){

				// This is where the contents of the parameter file are saved
				// We can write out this string at a later time
				if(this.completeFile == null){
					this.completeFile = line;
				}else{
					this.completeFile += "\n" + line;
				}

				// white space at the beginning and end of lines is ignored
				line = line.trim();
				
				// skip over comments (i.e., lines beginning with #) and malformed lines (i.e., lines without a colon)
				if(line.length() == 0 || line.charAt(0) == '#' || line.indexOf(':') < 0) continue;

				key = line.substring(0, line.indexOf(':')).toUpperCase().trim();
				value = line.substring(line.indexOf(':') + 1).trim();

				// key must contain at least one character
				if(key.length() > 0) this.table.put(key, value);
			}

		}catch(IOException e){
			System.err.println("(ConfigParams.ConfigParams) ERROR: An IOException was thrown while reading " + fileName);
		}
	}

	/**
	 * prints a copy of the parameter file to the location given by <code>fileName</code>
	 *
	 * @param fileName destination to copy the parameter file
	 */
	public void printToFile(final String fileName){

		try{
			final PrintWriter printWriter = new PrintWriter(new BufferedWriter(new FileWriter(fileName)));
			printWriter.println(this.completeFile);
			printWriter.close();
		}catch(IOException e){
			System.err.println("(ConfigParams.printParamsFile) ERROR: An IOException was thrown while writing " + fileName);
		}
	}

	/**
	 * Obtains the <code>int</code> value for a given parameter.
	 * Prints an error and returns <code>0</code> if the parameter doesn't exist or is not of type <code>int</code>.
	 *
	 * @param key parameter name
	 * @return value of parameter
	 */
	public int paramInt(String key){

		key = key.toUpperCase();

		if(!this.table.containsKey(key)){
			ConfigParams.printKeyNotFound("paramInt", key, "0");
			return 0;
		}

		try{
			return (new Integer((String)this.table.get(key))).intValue();
		}catch(NumberFormatException e){
			ConfigParams.printInvalidValueType("paramInt", key, "int", "0");
			return 0;
		}
	}

	/**
	 * Obtains the <code>long</code> value for a given parameter.
	 * Prints an error and returns <code>0</code> if the parameter doesn't exist or is not of type <code>long</code>.
	 *
	 * @param key parameter name
	 * @return value of parameter
	 */
	public long paramLong(String key){

		key = key.toUpperCase();

		if(!this.table.containsKey(key)){
			ConfigParams.printKeyNotFound("paramLong", key, "0");
			return 0;
		}

		try{
			return (new Long((String)this.table.get(key))).longValue();
		}catch(NumberFormatException e){
			ConfigParams.printInvalidValueType("paramLong", key, "long", "0");
			return 0;
		}
	}

	/**
	 * Obtains the <code>double</code> value for a given parameter.
	 * Prints an error and returns <code>0.0</code> if the parameter doesn't exist or is not of type <code>double</code>.
	 *
	 * @param key parameter name
	 * @return value of parameter
	 */
	public double paramDouble(String key){

		key = key.toUpperCase();

		if(!this.table.containsKey(key)){
			ConfigParams.printKeyNotFound("paramDouble", key, "0.0");
			return 0.0;
		}

		try{
			return (new Double((String)this.table.get(key))).doubleValue();
		}catch(NumberFormatException e){
			ConfigParams.printInvalidValueType("paramDouble", key, "double", "0.0");
			return 0.0;
		}
	}

	/**
	 * Obtains the <code>boolean</code> value for a given parameter.
	 * Prints an error and returns <code>false</code> if the parameter doesn't exist or is not of type <code>boolean</code>.
	 *
	 * @param key parameter name
	 * @return value of parameter
	 */
	public boolean paramBoolean(String key){

		key = key.toUpperCase();

		if(!this.table.containsKey(key)){
			ConfigParams.printKeyNotFound("paramBoolean", key, "false");
			return false;
		}

		try{
			return (new Boolean((String)this.table.get(key))).booleanValue();
		}catch(NumberFormatException e){
			ConfigParams.printInvalidValueType("paramBoolean", key, "boolean", "false");
			return false;
		}
	}

	/**
	 * Obtains the <code>String</code> value for a given parameter.
	 * Returns the empty string (i.e., <code>""</code>) if the parameter doesn't.
	 *
	 * @param key parameter name
	 * @return value of parameter
	 */
	public String paramString(String key){

		key = key.toUpperCase();

		if(!this.table.containsKey(key)){
			ConfigParams.printKeyNotFound("paramString", key, "\"\"");
			return "";
		}

		return (String)this.table.get(key);
	}

	/**
	 * Obtains the <code>Double2D</code> value for a given parameter.
	 * Prints an error and returns <code>new Double2D()</code> if the parameter doesn't exist or is not of type <code>Double2D</code>.
	 *
	 * @param key parameter name
	 * @return value of parameter
	 */
	public Double2D paramDouble2D(String key){

		key = key.toUpperCase();

		if(!this.table.containsKey(key)){
			ConfigParams.printKeyNotFound("paramDouble2D", key, "new Double2D()");
			return new Double2D();
		}


		try{
			final StringTokenizer t = new StringTokenizer((String)this.table.get(key));
			final double x = Double.parseDouble(t.nextToken());
			final double y = Double.parseDouble(t.nextToken());

			return new Double2D(x, y);
		}catch(NumberFormatException e){
			ConfigParams.printInvalidValueType("paramDouble2D", key, "Double2D", "new Double2D()");
			return new Double2D();
		}
	}

	/**
	 * Obtains an integer array for a given parameter.
	 * Prints an error and returns an array of size zero if the parameter doesn't exist or is not of type <code>int[]</code>.
	 *
	 * @param key parameter name
	 * @return value of parameter
	 */
	public int[] paramIntArray(String key){

		key = key.toUpperCase();

		if(!this.table.containsKey(key)){
			ConfigParams.printKeyNotFound("paramIntArray", key, "new int[]{}");
			return new int[]{};
		}

		try{
			final StringTokenizer t = new StringTokenizer((String)this.table.get(key));
			final int[] array = new int[t.countTokens()];

			for(int i = 0; i < array.length; i++){
				array[i] = Integer.parseInt(t.nextToken());
			}
			return array;
		}catch(NumberFormatException e){
			ConfigParams.printInvalidValueType("paramIntArray", key, "int array", "new int[]{}");
			return new int[]{};
		}
	}

	/**
	 * Obtains a double array for a given parameter.
	 * Prints an error and returns an array of size zero if the parameter doesn't exist or is not of type <code>double[]</code>.
	 *
	 * @param key parameter name
	 * @return value of parameter
	 */
	public double[] paramDoubleArray(String key){

		key = key.toUpperCase();

		if(!this.table.containsKey(key)){
			ConfigParams.printKeyNotFound("paramDoubleArray", key, "new double[]{}");
			return new double[]{};
		}

		try{
			final StringTokenizer t = new StringTokenizer((String)this.table.get(key));
			final double[] array = new double[t.countTokens()];

			for(int i = 0; i < array.length; i++){
				array[i] = Double.parseDouble(t.nextToken());
			}
			return array;
		}catch(NumberFormatException e){
			ConfigParams.printInvalidValueType("paramDoubleArray", key, "double array", "new double[]{}");
			return new double[]{};
		}
	}
	
	/**
	 * Obtains a boolean array for a given parameter.
	 * Prints an error and returns an array of size zero if the parameter doesn't exist or is not of type <code>boolean[]</code>.
	 *
	 * @param key parameter name
	 * @return value of parameter
	 */
	public boolean[] paramBooleanArray(String key){

		key = key.toUpperCase();

		if(!this.table.containsKey(key)){
			ConfigParams.printKeyNotFound("paramBooleanArray", key, "new boolean[]{}");
			return new boolean[]{};
		}

		try{
			final StringTokenizer t = new StringTokenizer((String)this.table.get(key));
			final boolean[] array = new boolean[t.countTokens()];

			for(int i = 0; i < array.length; i++){
				array[i] = t.nextToken().equalsIgnoreCase("true");
			}
			return array;
		}catch(Exception e){
			ConfigParams.printInvalidValueType("paramBooleanArray", key, "boolean array", "new boolean[]{}");
			return new boolean[]{};
		}
	}

	/**
	 * Obtains a String array for a given parameter.
	 * Prints an error and returns an array of size zero if the parameter doesn't exist or is not of type <code>String[]</code>.
	 *
	 * @param key parameter name
	 * @return value of parameter
	 */
	public String[] paramStringArray(String key){

		key = key.toUpperCase();

		if(!this.table.containsKey(key)){
			ConfigParams.printKeyNotFound("paramStringArray", key, "new String[]{}");
			return new String[]{};
		}

		try{
			final StringTokenizer t = new StringTokenizer((String)this.table.get(key));
			final String[] array = new String[t.countTokens()];

			for(int i = 0; i < array.length; i++){
				array[i] = t.nextToken();
			}
			return array;
		}catch(Exception e){
			ConfigParams.printInvalidValueType("paramStringArray", key, "String array", "new String[]{}");
			return new String[]{};
		}
	}

	/**
	 * Obtains a Color for a given parameter.
	 * Prints an error and returns <code>Color.black</code> if the parameter doesn't exist or is not of type <code>Color</code>.
	 * RGB colors are used with either three (Red, Green, Blue) or four (Red, Green, Blue, Alpha) parameters.
	 *
	 * @param key parameter name
	 * @return value of parameter
	 */
	public Color paramColor(String key){

		key = key.toUpperCase();

		if(!this.table.containsKey(key)){
			ConfigParams.printKeyNotFound("paramColor", key, "Color.black");
			return Color.black;
		}

		try{
			final StringTokenizer t = new StringTokenizer((String)this.table.get(key));

			if(t.countTokens() == 3){
				return new Color(Integer.parseInt(t.nextToken()), Integer.parseInt(t.nextToken()), Integer.parseInt(t.nextToken()));
			}else{
				return new Color(Integer.parseInt(t.nextToken()), Integer.parseInt(t.nextToken()), Integer.parseInt(t.nextToken()), Integer.parseInt(t.nextToken()));
			}
		}catch(Exception e){
			ConfigParams.printInvalidValueType("paramColor", key, "Color(int, int, int) or Color(int, int, int, int)", "Color.black");
			return Color.black;
		}
	}

	/**
	 * Prints a message to <code>System.err</code> indicating that the given key was not found in the parameter file.
	 */
	private static void printKeyNotFound(final String functionName, final String key, final String defaultValue){

		final String className = "ConfigParams";

		System.err.println("(" + className + "." + functionName + ") WARNING: " + key + " not found in parameter list. Default value '" + defaultValue + "' returned.");
	}

	/**
	 * Prints a message to <code>System.err</code> indicating that the value for the given key is not of the correct type.
	 */
	private static void printInvalidValueType(final String functionName, final String key, final String type, final String defaultValue){

		final String className = "ConfigParams";

		System.err.println("(" + className + "." + functionName + ") WARNING: Value for " + key + " is not of type " + type + ". Default value '" + defaultValue + "' returned.");
	}
}
